"""configuration service"""
from ast import literal_eval
from datetime import timedelta

from peewee import Database

from timeslime.models import Setting


class ConfigurationService():
    """class for the configuration service"""
    def __init__(self, database: Database):
        """create configuration service"""
        self._working_hours: list[timedelta] = []
        self._connection = database

        models = [Setting]
        self._connection.bind(models)
        self._connection.create_tables(models)

    def __str_to_timedelta(self, timedelta_string: str) -> timedelta:
        """convert string to timedelta"""
        timedelta_string_array = timedelta_string.split(':')
        return timedelta(
            hours=int(timedelta_string_array[0]),
            minutes=int(timedelta_string_array[1]),
            seconds=int(timedelta_string_array[2])
        )


    def _get_working_hours(self):
        setting = Setting.get(key="working_hours")
        self._working_hours = list(
            map(
                lambda item: self.__str_to_timedelta(item),
                literal_eval(setting.value)
            )
        )
        return self._working_hours

    def _set_working_hours(self, value: list[timedelta]):
        setting, _ = Setting.get_or_create(key="working_hours")
        setting.value = list(map(lambda item: str(item), value))
        setting.save()

        self._working_hours = value

    working_hours = property(
        fget=_get_working_hours,
        fset=_set_working_hours
    )

    def _get_timeslime_server(self):
        setting = Setting.get_or_none(key="timeslime_server")
        if setting is None:
            return None
        self._timeslime_server = setting.value
        return self._timeslime_server

    def _set_timeslime_server(self, value: str):
        setting, _ = Setting.get_or_create(key="timeslime_server")
        setting.value = value
        setting.save()

        self._working_hours = value

    timeslime_server = property(
        fget=_get_timeslime_server,
        fset=_set_timeslime_server
    )

    def _get_username(self):
        setting = Setting.get_or_none(key="username")
        if setting is None:
            return None
        self._timeslime_server = setting.value
        return self._timeslime_server

    def _set_username(self, value: str):
        setting, _ = Setting.get_or_create(key="username")
        setting.value = value
        setting.save()

        self._working_hours = value

    username = property(
        fget=_get_username,
        fset=_set_username
    )

    def _get_password(self):
        setting = Setting.get_or_none(key="password")
        if setting is None:
            return None
        self._timeslime_server = setting.value
        return self._timeslime_server

    def _set_password(self, value: str):
        setting, _ = Setting.get_or_create(key="password")
        setting.value = value
        setting.save()

        self._working_hours = value

    password = property(
        fget=_get_password,
        fset=_set_password
    )



"""synchronize service"""
from datetime import datetime, timezone
import logging
from typing import Optional
from urllib.parse import urljoin
from uuid import UUID

from peewee import Database
from requests import Session
from rich.console import Console

from timeslime.models import Setting, State, Timespan
from timeslime.services.configuration_service import ConfigurationService

class SynchronizeService():
    """class synchronize service"""
    def __init__(
            self,
            database: Database,
            configuration_service: ConfigurationService
        ):
        """create synchronize service"""
        self._connection = database
        self._configuration_service = configuration_service

        models = [Setting, State, Timespan]
        self._connection.bind(models)
        self._connection.create_tables(models)

        self.server_url = self._configuration_service.timeslime_server
        if not self.server_url:
            logging.error("Timeslime server is not configured!")

        self.timespan_route = urljoin(self.server_url, "api/v1/timespans")
        self.setting_route = urljoin(self.server_url, "api/v1/settings")

        self.session = Session()
        self.session.auth = (
            self._configuration_service.username,
            self._configuration_service.password
        )


    def sync(self, date: Optional[str] = None):
        """synchronize all data
        :param date: define since when the data should be synchronized"""
        console = Console()
        logging.info("Start synchronization")

        from_date = date
        if date is None:
            time_sync_date = State.get_or_none(key="time_sync_date")
            if not time_sync_date or from_date == "2000-01-01":
                from_date = datetime.now(
                        timezone.utc
                    ).astimezone().replace(
                        hour=0, minute=0, second=0, microsecond=0
                    ).strftime("%Y-%m-%d")
            else:
                from_date = time_sync_date.value

        response_json = self.__get_all_measurements(from_date)
        for measurement in response_json["data"]:
            start_time = None
            stop_time = None
            updated_at = None
            if "start_time" in measurement and measurement["start_time"] is not None:
                start_time = datetime.fromisoformat(measurement["start_time"])
            if "stop_time" in measurement and measurement["stop_time"] is not None:
                stop_time = datetime.fromisoformat(measurement["stop_time"])
            if "updated_at" in measurement and measurement["updated_at"] is not None:
                updated_at = datetime.fromisoformat(measurement["updated_at"])

            logging.debug("check if id: %s exists, if not add data", measurement["id"])
            timespan, created = Timespan.get_or_create(id=measurement["id"])
            if created:
                timespan.start_time = start_time
                timespan.stop_time = stop_time
                timespan.updated_at = updated_at
                timespan.save()
                logging.debug("Create timespan object: %s", timespan.id)
                continue

            logging.debug("check if data is newer on server, if yes update data")
            if timespan.updated_at < updated_at:
                timespan.start_time = start_time
                timespan.stop_time = stop_time
                timespan.updated_at = updated_at
                timespan.save()
                logging.debug("Update local timespan object: %s", timespan.id)
                continue

            logging.debug("check if data is newer local and send data to server")
            if timespan.updated_at != updated_at:
                logging.debug("Update remote timespan object: %s", timespan.id)
                self.__send_measurement(timespan)

        timespans = Timespan.select().where(
            Timespan.start_time > from_date
        )
        for timespan in timespans:
            remote_timespan = list(filter(lambda element: UUID(element["id"]) == timespan.id, response_json["data"]))
            if not remote_timespan:
                logging.info("Timespan object: %s is not on the server", timespan.id)
                self.__send_measurement(timespan)

        if from_date != "2000-01-01":
            time_sync_date, _ = State.get_or_create(key="time_sync_date")
            if time_sync_date.value != from_date:
                time_sync_date.value = from_date
                time_sync_date.save()
        logging.info("Finished synchronization")
        console.print("🤝 successfully synchronized your time and configuration data 🤝")

    def __get_all_measurements(self, from_date: str):
        """get all measurements from server"""

        logging.info("Get all data which are newer than date: %s", from_date)
        response = self.session.get(
            self.timespan_route,
            params={
                "from": from_date
            }
        )
        response.raise_for_status()

        response_json = response.json()
        logging.info("Received %s measurements", len(response_json["data"]))

        return response_json

    def __send_measurement(self, timespan: Timespan):
        """send measurement to server"""
        logging.info("Send timespan object to server: %s", timespan.id)
        start_time_str = None
        if timespan.start_time:
            start_time_str = str(timespan.start_time)

        stop_time_str = None
        if timespan.stop_time:
            stop_time_str = str(timespan.stop_time)

        response = self.session.post(self.timespan_route, json={
                "id": str(timespan.id),
                "start_time": start_time_str,
                "stop_time": stop_time_str,
                "updated_at": str(timespan.updated_at),
            }
        )
        response.raise_for_status()



"""report service"""
from datetime import datetime, timedelta, timezone
import logging
from typing import Optional

from peewee import Database
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table

from timeslime.models import Setting, Timespan
from timeslime.services.configuration_service import ConfigurationService

class ReportService():
    """class report service"""
    def __init__(
            self,
            database: Database,
            configuration_service: ConfigurationService
        ):
        """create report service"""
        self._connection = database
        self._configuration_service = configuration_service

        models = [Setting, Timespan]
        self._connection.bind(models)

    def report(self, date: Optional[str] = None, report_all: bool = False):
        """report time
        :param date: define since which date we want the report
        :param report_all: report report_all your time (default: False)"""
        logging.info("📊 report your time 📊")
        logging.debug("date: %s", date)
        logging.debug("report_all: %s", report_all)

        console = Console()
        if date and report_all:
            logging.error("Can't combine `--date` and `--all`")
            console.print(Markdown("Can't combine `--date` and `--all`\n\n"
                "You probably want to get all reports with:\n\n"
                "`timeslime report --all`\n\n"
                "or get the report for your specific date\n\n"
                f"`timeslime report --date={date}`"))
            return

        report_timespan_sum = timedelta()

        request_start_date = None
        request_stop_date = None
        if date:
            try:
                request_start_date = datetime.strptime(
                    date,
                    "%Y-%m-%d"
                ).astimezone()
            except ValueError:
                logging.error("`--date` has the wrong format.")
                today = datetime.now(
                    timezone.utc
                ).astimezone().replace(
                    hour=0, minute=0, second=0, microsecond=0
                ).strftime("%Y-%m-%d")
                console.print(Markdown("`--date` has the wrong format.\n\n"
                    "Allowed format is:\n\n"
                    "`timeslime report --date=YYYY-MM-DD`\n\n"
                    "Example for today:\n\n"
                    f"`timeslime report --date={today}`"))
                return
            request_stop_date = request_start_date + timedelta(days=1)
            logging.debug("request_start_date: %s", request_start_date)
            logging.debug("request_stop_date: %s", request_stop_date)
        else:
            request_start_date = datetime.now(
                    timezone.utc
                ).astimezone().replace(
                    hour=0, minute=0, second=0, microsecond=0
                )
            logging.debug("request_start_date: %s", request_start_date)

        table = Table(title="📊 timeslime report 📊", show_footer=True)

        table.add_column("🚀 Start Time", footer="Total")
        table.add_column("🛑 Stop Time")
        table.add_column("💪 Working time")

        output_format = "%H:%M:%S"
        if report_all:
            timespans = Timespan.select()
            output_format = "%Y-%m-%dT%H:%M:%S%z"
        else:
            if not request_stop_date:
                timespans = Timespan.select().where(
                    Timespan.start_time > request_start_date
                )
            else:
                timespans = Timespan.select().where(
                    Timespan.start_time >= request_start_date,
                    Timespan.start_time < request_stop_date,
                )

        # get configured time from today
        weekday = request_start_date.weekday()
        logging.debug("weekday: %s", weekday)
        target_working_time = self._configuration_service.working_hours[
            weekday
        ]

        for timespan in timespans:
            if timespan.stop_time:
                report_timespan_sum += timespan.stop_time - timespan.start_time
                table.add_row(
                    timespan.start_time.strftime(output_format),
                    timespan.stop_time.strftime(output_format),
                    str(timespan.stop_time - timespan.start_time),
                )
            else:
                now_with_timezone = datetime.now(timezone.utc).astimezone()
                difference = now_with_timezone - timespan.start_time
                report_timespan_sum += difference
                table.add_row(timespan.start_time.strftime(output_format), "-", str(difference))

        remaining_working_hours = target_working_time - report_timespan_sum
        table.columns[2].footer = str(abs(remaining_working_hours))
        if remaining_working_hours.total_seconds() < 0:
            table.footer_style = "green"
            table.columns[2].footer = str(f"+ {abs(remaining_working_hours)}")

        console.print(table)



"""collection of data models"""
from datetime import datetime, timezone
from uuid import uuid4

from peewee import DateTimeField, Model, TextField, UUIDField

DateTimeField.formats = [
    "%Y-%m-%d %H:%M:%S.%f%z",
    "%Y-%m-%d %H:%M:%S%z",
]


# pylint: disable=too-few-public-methods
class Setting(Model):
    """setting model"""
    id = UUIDField(primary_key=True, default=uuid4)
    key = TextField(index=True, null=True)
    value = TextField(null=True)

    class Meta:
        """defines meta information"""
        table_name = "settings"

class Timespan(Model):
    """timespan model"""
    id = UUIDField(primary_key=True, default=uuid4)
    start_time = DateTimeField(null=True)
    stop_time = DateTimeField(null=True)
    updated_at = DateTimeField(
        default=lambda: datetime.now(tz=timezone.utc), null=False
    )

    class Meta:
        """defines meta information"""
        table_name = "timespans"


class State(Model):
    """state model"""
    key = TextField(primary_key=True)
    value = TextField(null=True)

    class Meta:
        """defines meta information"""
        table_name = "state"
